<?php
/**
 * Ce fichier contient les filtres et les balises du plugin.
 */
if (!defined('_ECRIRE_INC_VERSION')) {
	return;
}

/**
 * Compile la balise `#FEED_CATEGORIES` qui fournit la configuration d'une ou des catégories affectées aux différents
 * feeds, pour un plugin utilisateur donné.
 * La signature de la balise est : `#FEED_CATEGORIES{plugin[, categorie_id, information]}`.
 *
 * @balise
 *
 * @example
 *     ```
 *     #FEED_CATEGORIES{isocode}, renvoie la description complète des catégories des feeds du plugin `isocode`.
 *     #FEED_CATEGORIES{isocode, map}, renvoie la description de la catégorie `map` des feeds du plugin `isocode`.
 *     #FEED_CATEGORIES{isocode, map, name}, renvoie le nom de la catégorie `map` des feeds du plugin `isocode`.
 *     ```
 *
 * @param Champ $p Pile au niveau de la balise.
 *
 * @return Champ Pile complétée par le code à générer.
 **/
function balise_FEED_CATEGORIES_dist(Champ $p) : Champ {
	// Récupération des arguments de la balise.
	// -- Préfixe du plugin (obligatoire)
	$plugin = interprete_argument_balise(1, $p);
	$plugin = isset($plugin) ? str_replace('\'', '"', $plugin) : '""';
	// -- Identifiant de la catégorie (facultatif)
	$categorie_id = interprete_argument_balise(2, $p);
	$categorie_id = isset($categorie_id) ? str_replace('\'', '"', $categorie_id) : '""';
	// -- Identifiant de la catégorie (facultatif)
	$information = interprete_argument_balise(3, $p);
	$information = isset($information) ? str_replace('\'', '"', $information) : '""';

	// Calcul de la balise
	$p->code = "(include_spip('inc/ezmashup_feed')
		? feed_repertorier_categories({$plugin}, {$categorie_id}, {$information})
		: [])";

	return $p;
}

/**
 * Compile la balise `#EZMASHUP_PLUGINS` qui fournit les plugins utilisateur actifs implémentant des feeds.
 * La signature de la balise est : `#EZMASHUP_PLUGINS{[actifs_seuls]}`.
 *
 * @balise
 *
 * @example
 *     ```
 *     #FEED_CATEGORIES, renvoie la liste des préfixes de plugins actifs utilisant Mashup Factory.
 *     #FEED_CATEGORIES{oui}, renvoie la liste des préfixes de plugins actifs utilisant Mashup Factory.
 *     #FEED_CATEGORIES{non}, renvoie la liste des préfixes de tous plugins utilisant Mashup Factory.
 *     ```
 *
 * @param Champ $p Pile au niveau de la balise.
 *
 * @return Champ Pile complétée par le code à générer.
 **/
function balise_EZMASHUP_PLUGINS_dist(Champ $p) : Champ {
	// Récupération des arguments de la balise.
	// -- Indicateur plugins actifs seulement ou tous
	$actifs_seuls = interprete_argument_balise(1, $p);
	$actifs_seuls = isset($actifs_seuls) ? str_replace('\'', '"', $actifs_seuls) : '"oui"';

	// Calcul de la balise
	$p->code = "ezmashup_lister_plugins({$actifs_seuls})";

	return $p;
}

/**
 * Renvoie la liste des préfixes des plugins actifs ou tous les plugins utilisant Mashup Factory.
 * L'utilisation du plugin Mashup Factory est repérée via les consignations dans la table `spip_meta`.
 *
 * @return array Liste des préfixes en minuscules
 */
function ezmashup_lister_plugins($actifs_seuls = 'oui') {
	static $plugins = [];

	// Eviter les valeurs erronées en rendant l'indicateur binarie
	$exclure_inactif = ($actifs_seuls === 'oui');

	if (!isset($plugins[$exclure_inactif])) {
		$prefixes = [];

		// On cherche les plugins utilisateurs à partir des métas de consignation
		$where = [
			'nom LIKE ' . sql_quote('%_ezmashup')
		];
		$metas = sql_allfetsel('nom', 'spip_meta', $where);
		if ($metas) {
			$metas = array_column($metas, 'nom');
			foreach ($metas as $_meta) {
				$prefixe = str_replace('_ezmashup', '', $_meta);
				if (
					(
						$exclure_inactif
						and defined('_DIR_PLUGIN_' . strtoupper($prefixe))
					)
					or !$exclure_inactif
				) {
					$prefixes[] = $prefixe;
				}
			}
		}

		// Enregistrement de la liste
		$plugins[$exclure_inactif] = $prefixes;
	}

	return $plugins[$exclure_inactif];
}

/**
 * Construit la liste des tags non techniques du feed pour un affichage.
 * Les tags non techniques commencent par un `_`.
 *
 * @param array       $tags       Tableau des tags du feed
 * @param null|string $separateur Séparateur des tags dans la liste affichée (par défaut une virgule)
 *
 * @return string Liste des tags non techniques au format chaine
 */
function feed_taglist_afficher(array $tags, ?string $separateur = ',') : string {
	// On exclut de l'affichage les tags techniques commençant par un '_'
	$affichage = '';
	$separateur = trim($separateur) . ' ';
	foreach ($tags as $_tag => $_valeur) {
		if (substr($_tag, 0, 1) !== '_') {
			$affichage .= ($affichage && $_valeur ? $separateur : '') . $_valeur;
		}
	}

	return $affichage;
}
